from pandas_datareader.iex import IEX
from datetime import datetime

# Data provided for free by IEX
# Data is furnished in compliance with the guidelines promulgated in the IEX
# API terms of service and manual
# See https://iextrading.com/api-exhibit-a/ for additional information
# and conditions of use


class Deep(IEX):
    """
    Retrieve order book data from IEX

    Notes
    -----
    Real-time depth of book quotations direct from IEX. Returns aggregated
    size of resting displayed orders at a price and side. Does not indicate
    the size or number of individual orders at any price level. Non-displayed
    orders and non-displayed portions of reserve orders are not counted.

    Also provides last trade price and size information. Routed executions
    are not reported.
    """
    def __init__(self, symbols=None, service=None, start=None, end=None,
                 retry_count=3, pause=0.001, session=None):
        if isinstance(symbols, str):
            symbols = symbols.lower()
        else:
            symbols = [s.lower() for s in symbols]

        super(Deep, self).__init__(symbols=symbols,
                                   start=start, end=end,
                                   retry_count=retry_count,
                                   pause=pause, session=session)
        self.sub = service

    @property
    def service(self):
        """Service endpoint"""
        ss = "/" + self.sub if self.sub is not None else ""
        return "deep{}".format(ss)

    def _read_lines(self, out):
        """
        IEX depth of book data varies and shouldn't always be returned in a DF

        :param out: Raw HTTP Output
        :return: DataFrame
        """

        # Runs appropriate output functions per the service being accessed.
        fmap = {
            'book': '_pass',
            'op-halt-status': '_convert_tstamp',
            'security-event': '_convert_tstamp',
            'ssr-status': '_convert_tstamp',
            'system-event': '_read_system_event',
            'trades': '_pass',
            'trade-breaks': '_convert_tstamp',
            'trading-status': '_read_trading_status',
            None: '_pass',
        }

        if self.sub in fmap:
            return getattr(self, fmap[self.sub])(out)
        else:
            raise "Invalid service specified: {}.".format(self.sub)

    def _read_system_event(self, out):
        # Map the response code to a string output per the API docs.
        # Per: https://www.iextrading.com/developer/docs/#system-event-message
        smap = {
            'O': 'Start of messages',
            'S': 'Start of system hours',
            'R': 'Start of regular market hours',
            'M': 'End of regular market hours',
            'E': 'End of system hours',
            'C': 'End of messages'
        }
        tid = out["systemEvent"]
        out["eventResponse"] = smap[tid]

        return self._convert_tstamp(out)

    @staticmethod
    def _pass(out):
        return out

    def _read_trading_status(self, out):
        # Reference: https://www.iextrading.com/developer/docs/#trading-status
        smap = {
            'H': 'Trading halted across all US equity markets',
            'O': 'Trading halt released into an Order Acceptance Period '
                 '(IEX-listed securities only)',
            'P': 'Trading paused and Order Acceptance Period on IEX '
                 '(IEX-listed securities only)',
            'T': 'Trading on IEX'
        }
        rmap = {
            # Trading Halt Reasons
            'T1': 'Halt News Pending',
            'IPO1': 'IPO/New Issue Not Yet Trading',
            'IPOD': 'IPO/New Issue Deferred',
            'MCB3': 'Market-Wide Circuit Breaker Level 3 - Breached',
            'NA': 'Reason Not Available',

            # Order Acceptance Period Reasons
            'T2': 'Halt News Dissemination',
            'IPO2': 'IPO/New Issue Order Acceptance Period',
            'IPO3': 'IPO Pre-Launch Period',
            'MCB1': 'Market-Wide Circuit Breaker Level 1 - Breached',
            'MCB2': 'Market-Wide Circuit Breaker Level 2 - Breached'
        }
        for ticker, data in out.items():
            if data['status'] in smap:
                data['statusText'] = smap[data['status']]

            if data['reason'] in rmap:
                data['reasonText'] = rmap[data['reason']]

            out[ticker] = data

        return self._convert_tstamp(out)

    @staticmethod
    def _convert_tstamp(out):
        # Searches for top-level timestamp attributes or within dictionaries
        if 'timestamp' in out:
            # Convert UNIX to datetime object
            f = float(out["timestamp"])
            out["timestamp"] = datetime.fromtimestamp(f/1000)
        else:
            for ticker, data in out.items():
                if 'timestamp' in data:
                    f = float(data["timestamp"])
                    data["timestamp"] = datetime.fromtimestamp(f/1000)
                    out[ticker] = data

        return out



from pandas_datareader.base import _BaseReader


class TSPReader(_BaseReader):

    """
    Returns DataFrame of historical TSP fund prices from symbols, over date
    range, start to end.

    Parameters
    ----------
    symbols : str, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : str, (defaults to '1/1/2010')
        Starting date, timestamp. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980')
    end : str, (defaults to today)
        Ending date, timestamp. Same format as starting date.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    """

    def __init__(self,
                 symbols=('Linc', 'L2020', 'L2030', 'L2040',
                          'L2050', 'G', 'F', 'C', 'S', 'I'),
                 start=None, end=None, retry_count=3, pause=0.001,
                 session=None):
        super(TSPReader, self).__init__(symbols=symbols,
                                        start=start, end=end,
                                        retry_count=retry_count,
                                        pause=pause, session=session)
        self._format = 'string'

    @property
    def url(self):
        """API URL"""
        return 'https://www.tsp.gov/InvestmentFunds/FundPerformance/index.html'

    def read(self):
        """ read one data from specified URL """
        df = super(TSPReader, self).read()
        df.columns = map(lambda x: x.strip(), df.columns)
        return df

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {'startdate': self.start.strftime('%m/%d/%Y'),
                'enddate': self.end.strftime('%m/%d/%Y'),
                'fundgroup': self.symbols,
                'whichButton': 'CSV'}

    @staticmethod
    def _sanitize_response(response):
        """
        Clean up the response string
        """
        text = response.text.strip()
        if text[-1] == ',':
            return text[0:-1]
        return text



from ftplib import FTP, all_errors
from pandas import read_csv
from pandas_datareader._utils import RemoteDataError
from pandas.compat import StringIO

import time
import warnings

_NASDAQ_TICKER_LOC = '/SymbolDirectory/nasdaqtraded.txt'
_NASDAQ_FTP_SERVER = 'ftp.nasdaqtrader.com'
_TICKER_DTYPE = [('Nasdaq Traded', bool),
                 ('Symbol', str),
                 ('Security Name', str),
                 ('Listing Exchange', str),
                 ('Market Category', str),
                 ('ETF', bool),
                 ('Round Lot Size', float),
                 ('Test Issue', bool),
                 ('Financial Status', str),
                 ('CQS Symbol', str),
                 ('NASDAQ Symbol', str),
                 ('NextShares', bool)]
_CATEGORICAL = ('Listing Exchange', 'Financial Status')

_DELIMITER = '|'
_ticker_cache = None


def _bool_converter(item):
    return item == 'Y'


def _download_nasdaq_symbols(timeout):
    """
    @param timeout: the time to wait for the FTP connection
    """
    try:
        ftp_session = FTP(_NASDAQ_FTP_SERVER, timeout=timeout)
        ftp_session.login()
    except all_errors as err:
        raise RemoteDataError('Error connecting to %r: %s' %
                              (_NASDAQ_FTP_SERVER, err))

    lines = []
    try:
        ftp_session.retrlines('RETR ' + _NASDAQ_TICKER_LOC, lines.append)
    except all_errors as err:
        raise RemoteDataError('Error downloading from %r: %s' %
                              (_NASDAQ_FTP_SERVER, err))
    finally:
        ftp_session.close()

    # Sanity Checking
    if not lines[-1].startswith('File Creation Time:'):
        raise RemoteDataError('Missing expected footer. Found %r' % lines[-1])

    # Convert Y/N to True/False.
    converter_map = dict((col, _bool_converter) for col, t in _TICKER_DTYPE
                         if t is bool)

    # For pandas >= 0.20.0, the Python parser issues a warning if
    # both a converter and dtype are specified for the same column.
    # However, this measure is probably temporary until the read_csv
    # behavior is better formalized.
    with warnings.catch_warnings(record=True):
        data = read_csv(StringIO('\n'.join(lines[:-1])), '|',
                        dtype=_TICKER_DTYPE, converters=converter_map,
                        index_col=1)

    # Properly cast enumerations
    for cat in _CATEGORICAL:
        data[cat] = data[cat].astype('category')

    return data


def get_nasdaq_symbols(retry_count=3, timeout=30, pause=None):
    """
    Get the list of all available equity symbols from Nasdaq.

    Returns
    -------
    nasdaq_tickers : pandas.DataFrame
        DataFrame with company tickers, names, and other properties.
    """
    global _ticker_cache

    if timeout < 0:
        raise ValueError('timeout must be >= 0, not %r' % (timeout,))

    if pause is None:
        pause = timeout / 3
    elif pause < 0:
        raise ValueError('pause must be >= 0, not %r' % (pause,))

    if _ticker_cache is None:
        while retry_count > 0:
            try:
                _ticker_cache = _download_nasdaq_symbols(timeout=timeout)
                retry_count = -1
            except RemoteDataError:
                # retry on any exception
                if retry_count <= 0:
                    raise
                else:
                    retry_count -= 1
                    time.sleep(pause)

    return _ticker_cache



from datetime import datetime

import pytest

import pandas_datareader.data as web


class TestIEXDaily(object):

    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    @property
    def start(self):
        return datetime(2015, 2, 9)

    @property
    def end(self):
        return datetime(2017, 5, 24)

    def test_iex_bad_symbol(self):
        with pytest.raises(Exception):
            web.DataReader("BADTICKER", "iex,", self.start, self.end)

    def test_iex_bad_symbol_list(self):
        with pytest.raises(Exception):
            web.DataReader(["AAPL", "BADTICKER"], "iex", self.start, self.end)

    def test_daily_invalid_date(self):
        start = datetime(2010, 1, 5)
        end = datetime(2017, 5, 24)
        with pytest.raises(Exception):
            web.DataReader(["AAPL", "TSLA"], "iex", start, end)

    def test_single_symbol(self):
        df = web.DataReader("AAPL", "iex", self.start, self.end)
        assert list(df) == ["open", "high", "low", "close", "volume"]
        assert len(df) == 578
        assert df["volume"][-1] == 19219154

    def test_multiple_symbols(self):
        syms = ["AAPL", "MSFT", "TSLA"]
        df = web.DataReader(syms, "iex", self.start, self.end)
        assert sorted(list(df)) == syms
        for sym in syms:
            assert len(df[sym] == 578)

    def test_multiple_symbols_2(self):
        syms = ["AAPL", "MSFT", "TSLA"]
        good_start = datetime(2017, 2, 9)
        good_end = datetime(2017, 5, 24)
        df = web.DataReader(syms, "iex", good_start, good_end)
        assert isinstance(df, dict)
        assert len(df) == 3
        assert sorted(list(df)) == syms

        a = df["AAPL"]
        t = df["TSLA"]

        assert len(a) == 73
        assert len(t) == 73

        expected1 = a.loc["2017-02-09"]
        assert expected1["close"] == 132.42
        assert expected1["high"] == 132.445

        expected2 = a.loc["2017-05-24"]
        assert expected2["close"] == 153.34
        assert expected2["high"] == 154.17

        expected3 = t.loc["2017-02-09"]
        assert expected3["close"] == 269.20
        assert expected3["high"] == 271.18

        expected4 = t.loc["2017-05-24"]
        assert expected4["close"] == 310.22
        assert expected4["high"] == 311.0



import pandas_datareader.data as web
from pandas_datareader.data import get_data_stooq


def test_stooq_dji():
    f = web.DataReader('^DJI', 'stooq')
    assert f.shape[0] > 0


def test_get_data_stooq_dji():
    f = get_data_stooq('^DAX')
    assert f.shape[0] > 0



from datetime import date

import numpy as np
import pandas as pd
import pandas.util.testing as tm
import pytest

import pandas_datareader.data as web
from pandas_datareader.exceptions import ImmediateDeprecationError


class TestGoogleOptions(object):

    @classmethod
    def setup_class(cls):
        cls.goog = None

    def test_deprecation(self):
        with pytest.raises(ImmediateDeprecationError):
            web.Options('GOOG', 'google')

    @pytest.mark.xfail(reason='Parsing error')
    def test_get_options_data(self):
        options = self.goog.get_options_data(expiry=self.goog.expiry_dates[0])

        assert isinstance(options, pd.DataFrame)
        assert len(options) > 10

        exp_columns = pd.Index(['Last', 'Bid', 'Ask', 'Chg',
                                'PctChg', 'Vol', 'Open_Int', 'Root',
                                'Underlying_Price', 'Quote_Time'])

        tm.assert_index_equal(options.columns, exp_columns)
        assert options.index.names == [u'Strike', u'Expiry',
                                       u'Type', u'Symbol']

        dtypes = ['float64'] * 7 + ['object', 'float64', 'datetime64[ns]']
        dtypes = [np.dtype(x) for x in dtypes]

        tm.assert_series_equal(options.dtypes, pd.Series(
            dtypes, index=exp_columns))

        for typ in options.index.levels[2]:
            assert typ in ['put', 'call']

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_options_data_yearmonth(self):
        with pytest.raises(ImmediateDeprecationError):
            self.goog.get_options_data(month=1, year=2016)

    @pytest.mark.xfail(reason='Parsing error')
    def test_expiry_dates(self):
        dates = self.goog.expiry_dates

        assert len(dates) == 1
        assert isinstance(dates, list)
        assert all(isinstance(dt, date) for dt in dates)

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_call_data(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_call_data()

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_put_data(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_put_data()

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_near_stock_price(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_near_stock_price()

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_forward_data(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_forward_data([1, 2, 3])

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_all_data(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_all_data()

    @pytest.mark.xfail(reason='Deprecated')
    def test_get_options_data_with_year(self):
        with pytest.raises(NotImplementedError):
            self.goog.get_options_data(year=2016)



import datetime as dt
import re
import json

from pandas import DataFrame, MultiIndex

from pandas_datareader.base import _OptionBaseReader


class Options(_OptionBaseReader):
    """
    ***Experimental***
    This class fetches call/put data for a given stock/expiry month.

    It is instantiated with a string representing the ticker symbol.

    The class has the following methods:
        get_options_data(month, year, expiry)
        get_call_data(month, year, expiry)
        get_put_data(month, year, expiry)
        get_near_stock_price(opt_frame, above_below)
        get_all_data(call, put)
        get_forward_data(months, call, put) (deprecated)

    Examples
    --------
    # Instantiate object with ticker
    >>> aapl = Options('aapl')

    # Fetch next expiry call data
    >>> calls = aapl.get_call_data()

    # Can now access aapl.calls instance variable
    >>> aapl.calls

    # Fetch next expiry put data
    >>> puts = aapl.get_put_data()

    # Can now access aapl.puts instance variable
    >>> aapl.puts

    # cut down the call data to be 3 below and 3 above the stock price.
    >>> cut_calls = aapl.get_near_stock_price(call=True, above_below=3)

    # Fetch call and put data with expiry from now to 8 months out
    >>> forward_data = aapl.get_forward_data(8, call=True, put=True)

    # Fetch all call and put data
    >>> all_data = aapl.get_all_data()
    """

    _OPTIONS_BASE_URL = "http://www.google.com/finance/option_chain?q={sym}" \
                        "&expd={day}&expm={month}&expy={year}&output=json"

    def get_options_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets call/put data for the stock with the expiration data in the
        given month and year

        Parameters
        ----------
        month : number, int, optional (default=None)
            The month the options expire. This should be either 1 or 2
            digits.

        year : number, int, optional (default=None)
            The year the options expire. This should be a 4 digit int.

        expiry : date-like or convertible or
                 list-like object, optional (default=None)
            The date (or dates) when options expire (defaults to current month)

        Returns
        -------
        pandas.DataFrame
            A DataFrame with requested options data.

            Index:
                Strike: Option strike, int
                Expiry: Option expiry, Timestamp
                Type: Call or Put, string
                Symbol: Option symbol as reported on Google, string
            Columns:
                Last: Last option price, float
                Chg: Change from prior day, float
                Bid: Bid price, float
                Ask: Ask price, float
                Vol: Volume traded, int64
                Open_Int: Open interest, int64
                Underlying: Ticker of the underlying security, string
                Underlying_Price: Price of the underlying security, float64
                Quote_Time: Time of the quote, Timestamp

        Notes
        -----
        Note: Format of returned data frame is dependent on
              Google and may change.

            >>> goog = Options('goog', 'google')  # Create object
            >>> goog.get_options_data(expiry=goog.expiry_dates[0])  # Get data

        """
        if month is not None or year is not None:
            raise NotImplementedError('month and year parameters '
                                      'cannot be used')
        if expiry is None:
            raise ValueError('expiry has to be set')
        d = self._load_data(expiry)
        return self._process_data(d, expiry)

    @property
    def expiry_dates(self):
        """
        Returns a list of available expiry dates
        """
        try:
            return self._expiry_dates
        except AttributeError:
            # has to be a non-valid date, to trigger returning 'expirations'
            d = self._load_data(dt.datetime(2016, 1, 3))
            self._expiry_dates = [dt.date(x['y'], x['m'], x['d'])
                                  for x in d['expirations']]
        return self._expiry_dates

    def _load_data(self, expiry):
        url = self._OPTIONS_BASE_URL.format(
            sym=self.symbol, day=expiry.day,
            month=expiry.month, year=expiry.year)
        s = re.sub(r'(\w+):', '"\\1":', self._read_url_as_StringIO(url).read())
        return json.loads(s)

    def _process_data(self, jd, expiry):
        """
        Parse JSON data into the DataFrame

        """
        now = dt.datetime.now()

        columns = ['Last', 'Bid', 'Ask', 'Chg', 'PctChg', 'Vol',
                   'Open_Int', 'Root', 'Underlying_Price', 'Quote_Time']
        indexes = ['Strike', 'Expiry', 'Type', 'Symbol']
        rows_list, index = self._process_rows(jd, now, expiry)
        df = DataFrame(rows_list, columns=columns,
                       index=MultiIndex.from_tuples(index, names=indexes))

        # Make dtype consistent, requires float64 as there can be NaNs
        df['Vol'] = df['Vol'].astype('float64')
        df['Open_Int'] = df['Open_Int'].astype('float64')

        return df.sort_index()

    def _process_rows(self, jd, now, expiry):
        rows_list = []
        index = []
        for key, typ in [['calls', 'call'], ['puts', 'put']]:
            for row in jd[key]:
                d = {}
                for dkey, rkey, ntype in [('Last', 'p', float),
                                          ('Bid', 'b', float),
                                          ('Ask', 'a', float),
                                          ('Chg', 'c', float),
                                          ('PctChg', 'cp', float),
                                          ('Vol', 'vol', int),
                                          ('Open_Int', 'oi', int)]:
                    try:
                        d[dkey] = ntype(row[rkey].replace(',', ''))
                    except (KeyError, ValueError):
                        pass
                d['Root'] = self.symbol
                d['Underlying_Price'] = jd['underlying_price']
                d['Quote_Time'] = now
                rows_list.append(d)
                index.append((float(row['strike'].replace(',', '')),
                              expiry, typ, row['s']))
        return rows_list, index



"""Custom warnings and exceptions"""


class UnstableAPIWarning(Warning):
    pass


DEP_ERROR_MSG = """
{0} has been immediately deprecated due to large breaks in the API without the
introduction of a stable replacement. Pull Requests to re-enable these data
connectors are welcome.

See https://github.com/pydata/pandas-datareader/issues
"""


class ImmediateDeprecationError(Exception):
    pass



from __future__ import unicode_literals

import pandas.compat as compat

from pandas_datareader.base import _BaseReader


class BankOfCanadaReader(_BaseReader):
    """Get data for the given name from Bank of Canada.

    Notes
    -----
    See `Bank of Canada <https://www.bankofcanada.ca/rates/>`__"""

    _URL = 'http://www.bankofcanada.ca/valet/observations'

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, compat.string_types):
            raise ValueError('data name must be string')

        return '{0}/{1}/csv'.format(self._URL, self.symbols)

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {'start_date': self.start.strftime('%Y-%m-%d'),
                'end_date': self.end.strftime('%Y-%m-%d')}

    @staticmethod
    def _sanitize_response(response):
        """
        Clean up the response string
        """
        data = response.text.split('OBSERVATIONS')[1]
        return data.split('ERRORS')[0].strip()



import pandas_datareader.data as web
from pandas_datareader._utils import RemoteDataError
from pandas_datareader._testing import skip_on_exception


class TestQuandl(object):
    # we test data from 10 years back where possible, 2 years otherwise, or...
    start10 = "2007-01-01"  # over ten years back
    end10 = "2007-01-05"
    day10 = "2007-01-04"
    start2 = "2015-01-01"  # over two years back
    end2 = "2015-01-05"
    day2 = "2015-01-02"

    def check_headers(self, df, expected_cols):
        expected_cols = frozenset(expected_cols)
        act_cols = frozenset(df.columns.tolist())
        assert expected_cols == act_cols, "unexpected cols: " + str(act_cols)

    @skip_on_exception(RemoteDataError)
    def test_db_wiki_us(self):
        df = web.DataReader('F', 'quandl', self.start10, self.end10)
        self.check_headers(df, ['Open', 'High', 'Low', 'Close', 'Volume',
                                'ExDividend', 'SplitRatio', 'AdjOpen',
                                'AdjHigh', 'AdjLow', 'AdjClose', 'AdjVolume'])
        assert df.Close.at[self.day10] == 7.70

    @skip_on_exception(RemoteDataError)
    def test_db_fse_frankfurt(self):
        # ALV_X: Allianz SE
        df = web.DataReader('FSE/ALV_X', 'quandl', self.start10, self.end10)
        self.check_headers(df, ['Open', 'High', 'Low', 'Close', 'Change',
                                'TradedVolume', 'Turnover',
                                'LastPriceoftheDay', 'DailyTradedUnits',
                                'DailyTurnover'])
        assert df.Close.at[self.day10] == 159.45

    @skip_on_exception(RemoteDataError)
    def test_db_sse_de_stuttgart(self):
        # ALV: Allianz SE
        df = web.DataReader('SSE/ALV', 'quandl', self.start2, self.end2)
        self.check_headers(df, [
                "High", "Low", "Last", "PreviousDayPrice", "Volume"])
        # as of 2017-06-11: PreviousDayPrice can be outside Low/High range;
        # Volume can be NaN
        assert df.Last.at[self.day2] == 136.47
        df2 = web.DataReader('ALV.DE', 'quandl', self.start2, self.end2)
        assert (df.Last == df2.Last).all()

    @skip_on_exception(RemoteDataError)
    def test_db_euronext_be_fr_nl_pt(self):
        # FP: Total SA
        # as of 2017-06-11, some datasets end a few months after their start,
        # e.g. ALVD, BASD
        df = web.DataReader('EURONEXT/FP', 'quandl', self.start2, self.end2)
        self.check_headers(df, [
                "Open", "High", "Low", "Last", "Turnover", "Volume"])
        assert df.Last.at[self.day2] == 42.525
        df2 = web.DataReader('FP.FR', 'quandl', self.start2, self.end2)
        assert (df.Last == df2.Last).all()

    @skip_on_exception(RemoteDataError)
    def test_db_lse_uk(self):
        # RBS: Royal Bank of Scotland
        df = web.DataReader('LSE/RBS', 'quandl', self.start10, self.end10)
        self.check_headers(df, ["High", "Low", "LastClose", "Price",
                                "Volume", "Change", "Var"])
        # as of 2017-06-11, Price == LastClose, all others are NaN
        assert df.Price.at[self.day10] == 5950.983

    @skip_on_exception(RemoteDataError)
    def test_db_nse_in(self):
        # TCS: Tata Consutancy Services
        df = web.DataReader('NSE/TCS', 'quandl', self.start10, self.end10)
        self.check_headers(df, ['Open', 'High', 'Low', 'Last', 'Close',
                                'TotalTradeQuantity', 'TurnoverLacs'])
        assert df.Close.at[self.day10] == 1259.05

    @skip_on_exception(RemoteDataError)
    def test_db_tse_jp(self):
        # TSE/6758: Sony Corp.
        df = web.DataReader('TSE/6758', 'quandl', self.start10, self.end10)
        self.check_headers(df, ['Open', 'High', 'Low', 'Close', 'Volume'])
        assert df.Close.at[self.day10] == 5190.0

    @skip_on_exception(RemoteDataError)
    def test_db_hkex_cn(self):
        # HKEX/00941: China Mobile
        df = web.DataReader('HKEX/00941', 'quandl', self.start2, self.end2)
        self.check_headers(df,
                           ['NominalPrice', 'NetChange', 'Change', 'Bid',
                            'Ask', 'PEx', 'High', 'Low', 'PreviousClose',
                            'ShareVolume000', 'Turnover000', 'LotSize'])
        assert df.High.at[self.day2] == 91.9



import re
import time
import warnings
import numpy as np
from pandas import Panel
from pandas_datareader.base import (_DailyBaseReader, _in_chunks)
from pandas_datareader._utils import (RemoteDataError, SymbolWarning)


class YahooDailyReader(_DailyBaseReader):

    """
    Returns DataFrame/Panel of historical stock prices from symbols, over date
    range, start to end. To avoid being penalized by Yahoo! Finance servers,
    pauses between downloading 'chunks' of symbols can be specified.

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, (defaults to '1/1/2010')
        Starting date, timestamp. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980')
    end : string, (defaults to today)
        Ending date, timestamp. Same format as starting date.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    adjust_price : bool, default False
        If True, adjusts all prices in hist_data ('Open', 'High', 'Low',
        'Close') based on 'Adj Close' price. Adds 'Adj_Ratio' column and drops
        'Adj Close'.
    ret_index : bool, default False
        If True, includes a simple return index 'Ret_Index' in hist_data.
    chunksize : int, default 25
        Number of symbols to download consecutively before intiating pause.
    interval : string, default 'd'
        Time interval code, valid values are 'd' for daily, 'w' for weekly,
        'm' for monthly and 'v' for dividend.
    """

    def __init__(self, symbols=None, start=None, end=None, retry_count=3,
                 pause=0.35, session=None, adjust_price=False,
                 ret_index=False, chunksize=25, interval='d'):
        super(YahooDailyReader, self).__init__(symbols=symbols,
                                               start=start, end=end,
                                               retry_count=retry_count,
                                               pause=pause, session=session,
                                               chunksize=chunksize)
        # Ladder up the wait time between subsequent requests to improve
        # probability of a successful retry
        self.pause_multiplier = 2.5

        self.headers = {
            'Connection': 'keep-alive',
            'Expires': str(-1),
            'Upgrade-Insecure-Requests': str(1),
            # Google Chrome:
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.99 Safari/537.36'  # noqa
        }

        self.adjust_price = adjust_price
        self.ret_index = ret_index
        self.interval = interval

        if self.interval not in ['d', 'wk', 'mo', 'm', 'w']:
            raise ValueError("Invalid interval: valid values are  'd', 'wk' and 'mo'. 'm' and 'w' have been implemented for "  # noqa
                             "backward compatibility. 'v' has been moved to the yahoo-actions or yahoo-dividends APIs.")  # noqa
        elif self.interval in ['m', 'mo']:
            self.pdinterval = 'm'
            self.interval = 'mo'
        elif self.interval in ['w', 'wk']:
            self.pdinterval = 'w'
            self.interval = 'wk'

        self.interval = '1' + self.interval
        self.crumb = self._get_crumb(retry_count)

    @property
    def service(self):
        return 'history'

    @property
    def url(self):
        return 'https://query1.finance.yahoo.com/v7/finance/download/{}'\
            .format(self.symbols)

    @staticmethod
    def yurl(symbol):
        return 'https://query1.finance.yahoo.com/v7/finance/download/{}'\
            .format(symbol)

    def _get_params(self, symbol):
        unix_start = int(time.mktime(self.start.timetuple()))
        day_end = self.end.replace(hour=23, minute=59, second=59)
        unix_end = int(time.mktime(day_end.timetuple()))

        params = {
            'period1': unix_start,
            'period2': unix_end,
            'interval': self.interval,
            'events': self.service,
            'crumb': self.crumb
        }
        return params

    def read(self):
        """ read one data from specified URL """
        try:
            df = super(YahooDailyReader, self).read()
            if self.ret_index:
                df['Ret_Index'] = _calc_return_index(df['Adj Close'])
            if self.adjust_price:
                df = _adjust_prices(df)
            return df.sort_index().dropna(how='all')
        finally:
            self.close()

    def _dl_mult_symbols(self, symbols):
        stocks = {}
        failed = []
        passed = []
        for sym_group in _in_chunks(symbols, self.chunksize):
            for sym in sym_group:
                try:
                    stocks[sym] = self._read_one_data(self.yurl(sym),
                                                      self._get_params(sym))
                    passed.append(sym)
                except IOError:
                    msg = 'Failed to read symbol: {0!r}, replacing with NaN.'
                    warnings.warn(msg.format(sym), SymbolWarning)
                    failed.append(sym)

        if len(passed) == 0:
            msg = "No data fetched using {0!r}"
            raise RemoteDataError(msg.format(self.__class__.__name__))
        try:
            if len(stocks) > 0 and len(failed) > 0 and len(passed) > 0:
                df_na = stocks[passed[0]].copy()
                df_na[:] = np.nan
                for sym in failed:
                    stocks[sym] = df_na
            return Panel(stocks).swapaxes('items', 'minor')
        except AttributeError:
            # cannot construct a panel with just 1D nans indicating no data
            msg = "No data fetched using {0!r}"
            raise RemoteDataError(msg.format(self.__class__.__name__))

    def _get_crumb(self, retries):
        # Scrape a history page for a valid crumb ID:
        tu = "https://finance.yahoo.com/quote/{}/history".format(self.symbols)
        response = self._get_response(tu,
                                      params=self.params, headers=self.headers)
        out = str(self._sanitize_response(response))
        # Matches: {"crumb":"AlphaNumeric"}
        rpat = '"CrumbStore":{"crumb":"([^"]+)"}'

        crumb = re.findall(rpat, out)[0]
        return crumb.encode('ascii').decode('unicode-escape')


def _adjust_prices(hist_data, price_list=None):
    """
    Return modifed DataFrame or Panel with adjusted prices based on
    'Adj Close' price. Adds 'Adj_Ratio' column.
    """
    if price_list is None:
        price_list = 'Open', 'High', 'Low', 'Close'
    adj_ratio = hist_data['Adj Close'] / hist_data['Close']

    data = hist_data.copy()
    for item in price_list:
        data[item] = hist_data[item] * adj_ratio
    data['Adj_Ratio'] = adj_ratio
    del data['Adj Close']
    return data


def _calc_return_index(price_df):
    """
    Return a returns index from a input price df or series. Initial value
    (typically NaN) is set to 1.
    """
    df = price_df.pct_change().add(1).cumprod()
    mask = df.iloc[1].notnull() & df.iloc[0].isnull()
    df.loc[df.index[0], mask] = 1

    # Check for first stock listings after starting date of index in ret_index
    # If True, find first_valid_index and set previous entry to 1.
    if (~mask).any():
        for sym in mask.index[~mask]:
            sym_idx = df.columns.get_loc(sym)
            tstamp = df[sym].first_valid_index()
            t_idx = df.index.get_loc(tstamp) - 1
            df.iloc[t_idx, sym_idx] = 1

    return df
